from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session
from requests.auth import HTTPBasicAuth
import requests
requests.packages.urllib3.disable_warnings() 

#Get Gas Token
def get_gas_token(client_id, client_secret, gas_token_url):
    """
    Gets the token from GAS that validates the user. For this project, 
    the user is the SFM technical user. Token needs to be given to SFM in
    order to receive an access token from SFM.

    Returns: 
    string
        the access token from GAS needed to be passed to SFM.
    """

    #function that fetches the token.
    auth = HTTPBasicAuth(client_id, client_secret)
    client = BackendApplicationClient(client_id=client_id)
    oauth = OAuth2Session(client=client)
    token = oauth.fetch_token(token_url=gas_token_url, auth=auth, headers={"Content-Type": "application/x-www-form-urlencoded"})
    return token['access_token']
#Format SFM Token Request
def get_SFM_token(client_id, client_secret, gas_token_url, SFM_token_url):
    """
    Gets the bearer token from SFM needed to authenticate API requests to SFM

    Returns:
    string
        returns the access token needed to authenticate future api requests
    """
    gas_token = get_gas_token(client_id, client_secret, gas_token_url)
    headers = {
        "Content-Type": "application/x-www-form-urlencoded"
    };
    creds = requests.post(url=SFM_token_url, headers=headers, data={'grant_type': 'token','access_token': gas_token}, verify=False)
    if creds.json()['access_token']:
        return creds.json()['access_token']
    else:
        raise Exception('No access token, bad techincal user credentials')


from typing import List, Union
import requests
class Measures:
    """
    Class containing methods to work with Measures through the SFM API
    """
    def get_measure_id(self, *object_ids: Union[str, List[str]]) -> List[str]:
        """
        Gets the measure id from SFM for a given object id in case you do
        not already have the measure id for the source you are getting values from. 
        This is needed because there is no api endpoint for trend data in SFM
        that takes object id as a paremeter, meaning you have to have the 
        measure id if you want trend data

        Args: 
            object_id (Union[str, List[str]]): The object id you want to get the measure_id for
            token (str): The token used to authenticate the api request
        
        Returns:
            measure_ids (List[str]): returns a list of measure ids
        """
        measure_ids = []
        for object_id in object_ids:
            measure_resp = requests.get(url=f"{self.sfm_url}/Measures", params={'dataobject_id': object_id}, headers=self.headers, verify=False).json()
            measure_id = measure_resp["Results"][0]['Id']
            measure_ids.append(measure_id)
        return measure_ids
    def update_measures(self, measures) -> object:
        
        """
        Creates a measure

        Args:
            measures: list of measure objects to create
            measure_group_id: Desired measure group to assign measure to
            title: desired title for the measure
            measure_type: Type of measure. Ex. 'Daily'
            description: Description of measure
            dependents (List[str], optional): list of the measure id's to include as dependents of this measure. Defaults to []
        """
        resp = {'success': [], 'failed': []}
        for measure in measures:
            params={
                "Title": measure['title'],
                "MeasureType": measure['measure_type'],
                "Definition": measure['definition'],
                "Description": measure['description'],
                "Source": measure['source'],
                "ValueSource": "External",
                "TargetSource": "None",
                "Category": measure['category'],
                "ShowLastValidValue": measure['show_last_valid_value'],
                "BaseType": measure['base_type'],
                "AggregateFunction": measure['aggregate_func'],
                "Dependents": measure['dependents'] if 'dependents' in measure else [],
                "TileDisplayType": "StandardTile",
                "TrendDisplayType": measure['trend_display_type'],
                "TargetType": measure['target_type'],
                "TargetBoundaries": measure['target_boundaries']
            }
            created_measure = requests.put(url=f"{self.sfm_url}/Measures/{measure['measure_id']}", json=params, headers=self.headers, verify=False)
            if created_measure.status_code <300:
                #201 is created
                resp['success'].append(created_measure.json())
            else:
                resp['failed'].append(created_measure.json())
        return resp
    
    def create_measure(self, measure) -> object:
        params={
                "Title": measure['title'],
                "MeasureType": measure['measure_type'],
                "Definition": measure['definition'],
                "Description": measure['description'],
                "Source": measure['source'],
                "ValueSource": "External",
                "TargetSource": "None",
                "Category": measure['category'],
                "ShowLastValidValue": measure['show_last_valid_value'],
                "BaseType": measure['base_type'],
                "AggregateFunction": measure['aggregate_func'],
                "Dependents": measure['dependents'] if 'dependents' in measure else [],
                "TileDisplayType": "StandardTile",
                "TrendDisplayType": measure['trend_display_type'],
                "TargetType": measure['target_type'],
                "TargetBoundaries": measure['target_boundaries']
            }
        return requests.post(url=f"{self.sfm_url}/MeasureGroups/{measure['measure_group_id']}/Measures", json=params, headers=self.headers, verify=False)
    
    def create_measures(self, measures) -> object:
        
        """
        Creates a measure

        Args:
            measures: list of measure objects to create
            measure_group_id: Desired measure group to assign measure to
            title: desired title for the measure
            measure_type: Type of measure. Ex. 'Daily'
            description: Description of measure
            dependents (List[str], optional): list of the measure id's to include as dependents of this measure. Defaults to []
        """
        resp = {'success': [], 'failed': []}
        for measure in measures:
            params={
                "Title": measure['title'],
                "MeasureType": measure['measure_type'],
                "Definition": measure['definition'],
                "Description": measure['description'],
                "Source": measure['source'],
                "ValueSource": "External",
                "TargetSource": "None",
                "Category": measure['category'],
                "ShowLastValidValue": measure['show_last_valid_value'],
                "BaseType": measure['base_type'],
                "AggregateFunction": measure['aggregate_func'],
                "Dependents": measure['dependents'] if 'dependents' in measure else [],
                "TileDisplayType": "StandardTile",
                "TrendDisplayType": measure['trend_display_type'],
                "TargetType": measure['target_type'],
                "TargetBoundaries": measure['target_boundaries']
            }
            created_measure = requests.post(url=f"{self.sfm_url}/MeasureGroups/{measure['measure_group_id']}/Measures", json=params, headers=self.headers, verify=False)
            if created_measure.status_code <300:
                #201 is created
                resp['success'].append(created_measure.json())
            else:
                resp['failed'].append(created_measure.json())
        return resp

    def delete_measure(self, measure_group_id: str, measure_id: str) -> bool:
        
        if (self.DEBUG):
            print(f"Deleting measure '{measure_id}' in measure group '{measure_group_id}'")
        resp = requests.delete(
            url=f"{self.sfm_url}/MeasureGroups/{measure_group_id}/Measures",
            params={'measureId': measure_id},
            headers=self.headers, 
            verify=False
        )
        
        if resp.status_code>=400:
            
            if (self.DEBUG):
                print(f"failed first time to delete measure '{measure_id}' in measure group '{measure_group_id}'. Trying to remove parents from current measure", resp.json())
            parents = requests.get(
                url=f"{self.sfm_url}/Measures/{measure_id}/Parents",
                headers=self.headers, 
                verify=False
            ).json()
            for parent in parents:
                # convert to function and add mutex?
                parent_resp = requests.get(
                    url=f"{self.sfm_url}/Measures/{parent['Id']}",
                    headers=self.headers, 
                    verify=False
                )
                if parent_resp.status_code == 500:
                    self.delete_measure(parent['MeasureGroupId'], parent['Id'])
                    parent_resp = requests.get(
                        url=f"{self.sfm_url}/Measures/{parent['Id']}",
                        headers=self.headers, 
                        verify=False
                    )
                
                m = parent_resp.json()
                
                if len(m['Dependents'])>1:
                    m['Dependents'] = [x['Id'] for x in m['Dependents'] if x['Id'] != measure_id]
                else:
                    m['ValueSource'] = 'External' if m['ValueSource']=='Aggregated' else m['ValueSource']
                    m['TargetSource'] = 'External' if m['TargetSource']=='Aggregated' else m['ValueSource']
                requests.put(
                    url=f"{self.sfm_url}/Measures/{parent['Id']}",
                    data=m,
                    headers=self.headers, 
                    verify=False
                ).json()
            resp = requests.delete(
                url=f"{self.sfm_url}/MeasureGroups/{measure_group_id}/Measures",
                params={'measureId': measure_id},
                headers=self.headers, 
                verify=False
            )
        return 200 <= resp.status_code < 300

    def clean_measure_group(self, measure_group_id: str, keep: list = [], keep_property: str = 'Id'):
        
        measures = requests.get(
            url=f"{self.sfm_url}/Measures",
            params={'measureGroupId': measure_group_id},
            headers=self.headers, 
            verify=False
        ).json()
        rerun = measures['IsResultsLimited']
        measures = [x for x in measures['Results'] if x[keep_property] not in keep]
        measures = sorted(measures, key = lambda x: x['ValueSource'])
        if (self.DEBUG):
            print(f"Deleting {len(measures)} measures in group '{measure_group_id}'")
        resp = {'success': 0, 'failed':0, 'errors': []}
        for measure in measures:
            delete_check = self.delete_measure(measure_group_id, measure['Id'])
            if delete_check:
                resp['success']+=1
            else:
                resp['failed']+=1
        if (rerun):
            sub_resp = self.clean_measure_group(measure_group_id, keep, keep_property)
            resp['success']+=sub_resp['success']
            resp['failed']+=sub_resp['failed']
            resp['errors']+=sub_resp['errors']
        return resp
    
    def delete_measures_in_group(self, measure_group_id: str, delete_list: list = []):
        pass
    
    def delete_measures(self, measures: list = []):
        
        resp = {'success': 0, 'failed': 0, 'errors': []}
        for measure in measures:
            check_delete = self.delete_measure(measure['kpi_group_id'],measure['measure_id'])
            if check_delete:
                resp["success"]+=1
            else:
                resp['failed']+=1
                


from typing import List
import requests
import asyncio
import concurrent.futures

class MeasureGroups:
    def get_measure_groups(self,forum_id: str):
        """
        Function to grab all measure groups for a forum structure

        Parameters:
            forum_id: forum structure to get measures from
        """
        api_resp = requests.get(
                        url=f"{self.sfm_url}/ForumStructures/{forum_id}/MeasureGroups", 
                        headers=self.headers, 
                        verify=False
                    )
        if api_resp.status_code == 200:
            return api_resp.json()
        else:
            return api_resp
    def create_measure_groups(self,forum_id: str, names: List[str]):
        """
        Checks if a measure group with name already exists. 
        If not, it makes a measure group with the specified name on the specified forum structure

        Paramters: 
            forum_id: the forum_id to create the measure group for
            names: the names of each measure group you want to create
        Returns:
            {success: [], failed: [], already_exist: []}
        
        """
        resp = {'success': [], 'failed':[], 'already_exist': []}
        try:
            measure_groups = requests.get(
                url=f"{self.sfm_url}/ForumStructures/{forum_id}/MeasureGroups", 
                headers=self.headers, 
                verify=False
            ).json()
            for name in names:
                temp = False
                for measure_group in measure_groups:
                    if measure_group['GroupName'] == name:
                        temp = True
                if not temp:
                    api_resp = requests.put(
                        url=f"{self.sfm_url}/ForumStructures/{forum_id}/MeasureGroups", 
                        params={'name': name}, 
                        headers=self.headers, 
                        verify=False
                    )
                    if api_resp.status_code < 300:
                        resp['success'].append(api_resp.json())
                    else:
                        resp['failed'].append(api_resp.json())
                else:
                    resp['already_exist'].append(name)
            if len(resp['already_exist']) > 0:
                print('These names already exist', resp['already_exist'])
        except Exception as e :
            print(e)
        return resp
    
    
    def delete_measure_groups(self, measure_groups: List[str] = []):
        resp = {'success': 0, 'failed':0, 'errors': []}
        for measure_group_id in measure_groups:
            try:
                api_resp = self.delete_measure_group(measure_group_id)

                if (200 == api_resp.status_code):
                    resp['success']+=1
                else:
                    raise Exception({'status_code': api_resp.status_code,'resp':api_resp.json()})
            except Exception as e:
                resp['failed']+=1
                resp['errors'].append(e)
        return resp
    
    def delete_measure_group(self, measure_group_id: str):
        
        if (self.DEBUG):
            print(f"Deleting measure group '{measure_group_id}'")
        api_resp = requests.delete(
            url=f"{self.sfm_url}/MeasureGroups/{measure_group_id}",
            headers=self.headers, 
            verify=False
        )
        if (api_resp.status_code in (400, 500)):
            self.clean_measure_group(measure_group_id)
            api_resp = requests.delete(
                url=f"{self.sfm_url}/MeasureGroups/{measure_group_id}",
                headers=self.headers, 
                verify=False
            )
        if (self.DEBUG):
            print('delete measure group resp',api_resp.json())
        return api_resp
    
    def move_measures(self, measure_group_id: str, measures: List[str] = []):
        
        if (self.DEBUG):
            print(f"Moving {len(measures)} measures into group {measure_group_id}")
        api_resp = requests.put(
            url=f"{self.sfm_url}/MeasureGroups/{measure_group_id}/Measures",
            json=measures,
            headers=self.headers,
            verify=False
        )
        if (self.DEBUG):
            print('move measure resp', api_resp.json())
        return api_resp

    
    # def delete_measure_groups(self, measure_groups: List[str] = []):
    #     resp = {'success': 0, 'failed':0, 'errors': []}
        
    #     async def looper(measure_groups):
    #         with concurrent.futures.ThreadPoolExecutor() as executor:
                
    #             loop = asyncio.get_event_loop()
    #             futures = [loop.run_in_executor(executor, self.delete_measure_group, measure_group_id) for measure_group_id in measure_groups]
    #             for response in await asyncio.gather(*futures):
    #                 try:
    #                     if (200 == response.status_code):
    #                         resp['success']+=1
    #                     else:
    #                         raise Exception({'status_code': response.status_code,'resp':response.json()})
    #                 except Exception as e:
    #                     resp['failed']+=1
    #                     resp['errors'].append(e)
                    
    #     asyncio.run(looper(measure_groups))
    #     return resp
    
    # def delete_measure_group(self, measure_group_id: str):
    #     if (self.DEBUG):
    #         print(f"Deleting measure group '{measure_group_id}'")
    #     api_resp = requests.delete(
    #         url=f"{self.sfm_url}/MeasureGroups/{measure_group_id}",
    #         headers=self.headers, 
    #         verify=False
    #     )
        
    #     if (400 == api_resp.status_code):
    #         if (self.DEBUG):
    #             print(api_resp.json())
    #         self.clean_measure_group(measure_group_id)
    #         api_resp = requests.delete(
    #             url=f"{self.sfm_url}/MeasureGroups/{measure_group_id}",
    #             headers=self.headers,
    #             verify=False
    #         )
    #     if (self.DEBUG):
    #         print(api_resp.json())
    #     return api_resp


import requests
from datetime import datetime
class ExternalValues:
    def upload_external_values(self, external_values, measure_type, measure_date = datetime.isoformat(datetime.now())):
        """
        Function to put external values 

        Parameters:
            external_values: array of external values to put
        """
        api_resp = requests.put(
                        url=f"{self.sfm_url}/Integration/ExternalValues", 
                        headers=self.headers, 
                        json=external_values,
                        params={'measureDate': measure_date, 'measureType': measure_type},
                        verify=False
                    )
        if api_resp.status_code < 300:
            return api_resp.json()
        else:
            raise Exception('Error with uploading: ', api_resp.json())
    


from mercedesSFMConnector.api.measure_groups import MeasureGroups
from mercedesSFMConnector.api.measures import Measures
from mercedesSFMConnector.api.external_values import ExternalValues
from mercedesSFMConnector.token_handler import TokenHandler
import os
import requests

#purpose is to handle api calls
class Api(Measures, MeasureGroups, ExternalValues):
    """
    Class to handle API calls to SFM

    Attributes:
        sfm_url (str): Base api url for SFM
        headers(obj): Headers for the api requests
    
    Methods: 
        generic_call
    """
    def __init__(
        self, 
        sfm_url: str = os.getenv('SFM_URL'), 
        gas_url: str = os.getenv('GAS_URL'), 
        client_id: str = os.getenv('SFM_TECH_USER_CLIENT_ID'), 
        client_secret: str = os.getenv('SFM_TECH_USER_CLIENT_SECRET')
        ) -> None:
        """
        __init__

        Args:
            sfm_url: BASE url for SFM. Ex. https://base-sfm-url.corpintra.net
            gas_url: BASE url for GAS.
            client_id: technical client id for SFM
            client_secret: technical client secret for SFM
        """
        if sfm_url == None or gas_url == None or client_id == None or client_secret == None:
            raise ValueError("Please specify parameters or place them in your .env file")
        self.__token_handler = TokenHandler(
            gas_url=gas_url, 
            sfm_url=sfm_url, 
            client_id=client_id, 
            client_secret=client_secret
            )
        self.sfm_url = f"{sfm_url}/rest/api"
        self.headers = {
            'Authorization': f"Bearer {self.__token_handler.access_token}"
        }
        
    def generic_call(self, endpoint: str, params:object = {}, body: object = {}, method: str ='GET') -> object:
        """
        Method to allow api calls if library ever outdated. Allows you to specify custom endpoints, methods, bodies, etc.

        Args:
            endpoint: the endpoint you wish to request
            params: query parameters 
            body: body for the request
            method: request method

        Returns:
            API response from query
        """
        api_resp = {}
        if method == 'GET':
            api_resp = requests.get(url=f"{self.sfm_url}/{endpoint}", params=params, headers=self.headers, verify=False).json()
        elif method == 'POST':
            api_resp = requests.post(url=f"{self.sfm_url}/{endpoint}", params=params, data=body, headers=self.headers, verify=False).json()

        return api_resp




from setuptools import find_packages, setup
setup(
    name='mercedesSFMConnector',
    packages=find_packages(include=['mercedesSFMConnector', 'mercedesSFMConnector.api', 'mercedesSFMConnector.token_handler']),
    version='0.1.4',
    description='A library to interact with the Mercedes SFM API in Python',
    author='Ethan Wilkes',
    install_requires=['requests', 'requests-oauthlib','oauthlib'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest==4.4.1'],
    test_suite='tests',
    license='MIT',
)



from msilib.schema import Error
import os
import mercedesSFMConnector.token_handler.token_libs as libs 
    
class TokenHandler:
    """
    Class designed to handle gas and sfm token

    Attributes:
        gas_token_url: url used to get GAS token
        sfm_token_url: url used to get SFM token

    """
    def __init__(
        self, 
        gas_url: str = os.getenv('GAS_URL'), 
        sfm_url: str = os.getenv('SFM_URL'),
        client_id: str = os.getenv('SFM_TECH_USER_CLIENT_ID'),
        client_secret: str = os.getenv('SFM_TECH_USER_CLIENT_SECRET')
        ) -> None:
        self.gas_token_url = f"{gas_url}/as/token.oauth2"
        self.sfm_token_url = f"{sfm_url}/rest/api/token"
        self.__client_id = client_id
        self.__client_secret = client_secret
        self.generate_access_token()
    
    @property
    def access_token(self):
        """
        str: Access token required in header in order to access SFM
        """
        return self.__access_token

    @access_token.setter
    def access_token(self, _value):
        raise Exception('Cannot set access_token. Run generate_access_token to generate a new token.')

    def generate_access_token(self):
        """
        Generates new SFM token. Used if token expires
        """
        self.__access_token = libs.get_SFM_token(
            client_id=self.__client_id, 
            client_secret=self.__client_secret,
            gas_token_url=self.gas_token_url,
            SFM_token_url=self.sfm_token_url
        )
